/*
 *  Steampunk hydraulic clock
 *
 *  Version 1.0    11/01/2019
 *
 *  This software drives an electric pump that operates a 1-minute siphon, which in
 *  turn operates a clock movement.  The pump continuously fills a cylindrical
 *  container that, once filled, drains via a siphon.  A float in the cylinder
 *  connected to a pawl advances a ratchet wheel 1/60th of a revolution to drive
 *  the clock movement.  A Hall effect sensor connected to the microcontroller
 *  generates an interrupt each time the siphon empties. The software monitors
 *  the cycle time and adjusts the duration of a pulse of current to the pump to
 *  maintain a precise 1 minute cycle time for the siphon.
 *
 *  Timer A2 runs continuously and generates an interrupt every 100 mS. This is used
 *  as a timebase to regulate the pump speed and the accuracy of the clock. The timer
 *  also generates a modulated pulse to the pump to regulate its speed to ensure that
 *  the cycle time to fill and drain the siphon is exactly one minute.
 *
 *
 *  Dick Bipes
 *  dick@carveshop.com
 *
 *
 *  (c) Copyright 2019 by Dick Bipes  All rights reserved
 *
 *
 */
/*
 * ======== Standard MSP430 includes ========
 */
#include <msp430.h>

/*
 * ======== Grace related includes ========
 */
#include <ti/mcu/msp430/csl/CSL.h>
//	Port 1 I/O
#define RedLED			(BIT0)		// Red LED on the LaunchPad
#define GreenLED		(BIT6)		// Green LED on the LaunchPad
#define RedLED2			(BIT3)		// Red LED
#define GreenLED2		(BIT4)		// Green LED
//						            // P1.2 is the timer PWM drive to the MOSFET pump
#define HallSensor		(BIT1)		// Hall sensor that triggers when the float is low

#define	TicksPerCycle	600		// number of timer ticks per desired 1-minute cycle

// These constants are determined by pump characteristics, plumbing, and cylinder size
#define NominalPWM	1500	// Nominal PWM comparator value - 28 mS 10 Hz
#define MinPWM		800		// Minimum PWM value to ensure that the pump fills
							// the upper cylinder
#define MaxPWM		3000	// Maximum PWM value to ensure that the pump does not
							// overpower the siphon
#define Kp	.8   			// Proportional constant, empirically derived
#define Ki	.4 				// Integral constant, empirically derived

unsigned int 	ticktock = 0;		// Timer tick - one count per 100 ms
unsigned int 	timer_capture;		// current value of ticktock
unsigned int 	last_capture = 0;	// previous value of ticktock

int error = 0;		// error, the difference between reference time base and
					//the siphon (clock)
int i_error = 0;	// integral of the error

#define	ArraySize 15
int errorhist[ArraySize] = {0};		// log the error history for development/debug
int i = 0;							// pointer to next entry in the error log
int	error_counter = 0;				// number of seconds the last cycle was in error
int i_error_counter = 0;			// integral of error - the total number of seconds
									// the clock is in error since it started running

int seconds10x = 0;
int minutes = 0;
int hours = 0;


int pumpPWM = NominalPWM;	// duration of the power pulse to the motor
							//	(sets the pump speed)

/*
 * Routine to wait a bit while an LED is turned on or off
 */
   void WaitABit(void)
   {
	   volatile unsigned int i;		// volatile to prevent optimization
	   for (i=0; i<30000; i++);		// wait a while
   }

/*
 * Interrupt service routine called when the Hall Sensor indicates that the float
 * is low.
 */
   void FloatLow(void)
   {
		P1IFG &= ~HallSensor; 				// Clear the interrupt
		timer_capture = ticktock;			// capture the time for the last fill/drain cycle
											// compute the error for this last cycle
		error = TicksPerCycle - (timer_capture - last_capture);
		last_capture = timer_capture;		// remember the timer count for next cycle
		i_error += error;					// integral of the error
		error_counter = error/10;			// copy error to count LED flashes
		i_error_counter = i_error/10;		// copy i_error to count LED flashes

		errorhist[i++] = error;				// keep a log of error
		if (i >= ArraySize)
			i = 0;

		pumpPWM = pumpPWM - (Kp*error + Ki*i_error);	// Standard PID algorithm, but with no
														// differential term

		if (pumpPWM < MinPWM)			// keep the PWM in bounds
			pumpPWM = MinPWM;			// long enough to ensure some water is pumped
		if (pumpPWM > MaxPWM)
			pumpPWM = MaxPWM;			// short enough to prevent overpowering the siphon

		TACCR1 = pumpPWM;				// set the timer to enable the new PWM time

   }
/*
* Interrupt service routine called when the timer reaches max value - once
* every 100 mS
*/
   void TimerTick(void)
   {
	   ticktock++;
	   // For debug, keep track of how long the clock was running
	   seconds10x++;
	   if (seconds10x >= 600)
	   {
		   minutes++;
		   seconds10x = 0;
	   }
	   if (minutes >= 60)
	   {
		   hours++;
		   minutes = 0;
	   }
   }

/*
 * 		Flash LEDs to indicate the number of seconds that the cycle deviates
 * 		from the ideal
 */
   void FlashI_error(void)
   {
	   while ((error_counter != 0) || (i_error_counter !=0) )
	   {
		   WaitABit();

		   if (error_counter > 0)		// seconds that the last cycle differs from 60 seconds
		   {
			   P1OUT |= GreenLED;		// turn the LED on
			   error_counter--;			// count down the number of seconds to zero
		   }
		   if (error_counter < 0)
		   {
			   P1OUT |= RedLED;			// turn the LED on
			   error_counter++;			// count up the number of seconds to zero
		   }

		   if (i_error_counter > 0)		// seconds that the clock deviates from ideal time
		   {
			   P1OUT |= GreenLED2;		// turn the LED on
			   i_error_counter--;		// count down the number of seconds to zero
		   }
		   if (i_error_counter < 0)
		   {
			   P1OUT |= RedLED2;		// turn the LED on
			   i_error_counter++;		// count up the number of seconds to zero
		   }

		   WaitABit();
		   P1OUT &= ~(GreenLED + RedLED + GreenLED2 + RedLED2);	// turn the LEDs off
	   }
   }

/*
 *  ======== main ========
 */
int main(int argc, char *argv[])
{
    CSL_init();             // Activate Grace-generated configuration
    
    // >>>>> Fill-in user code here <<<<<
    
    TACCR1 = NominalPWM;	// set the starting pump PWM duration


    for ( ; ; )
    {
    	FlashI_error();		// display the error times
    }

 }
